==> hqq-0.1.7.tar.gz: hqq-0.1.7/tests/test_quantize.py <==
import unittest, tempfile
import torch
from hqq.core.quantize import Quantizer, HQQLinear, BaseQuantizeConfig, HQQBackend
from hqq.engine.hf import HQQModelForCausalLM
from transformers import AutoModelForCausalLM, AutoConfig

# python -m unittest tests.test_quantize.TestQuantizer.test_quantizer
class TestQuantizer(unittest.TestCase):

    def setUp(self) -> None:
        # set seed
        torch.manual_seed(42)
        self.m = torch.nn.Linear(16,128)
        self.w = self.m.weight.data
        # create tmp dir
        self.tmp_dir = tempfile.mkdtemp()
        return super().setUp()
    
    def tearDown(self) -> None:
        return super().tearDown()
    
    def test_quantizer(self):    
        for compute_dtype in [torch.float32, torch.float16, torch.bfloat16]:
            for nbits in [8,4,3,2,1]:
                W_q, meta = Quantizer.quantize(self.w, nbits=nbits, round_zero=True, optimize=True, view_as_float=False)
                if nbits == 3:
                    assert W_q.dtype == torch.int32
                else:
                    assert W_q.dtype == torch.uint8
                w_dq = Quantizer.dequantize(W_q, meta)
                norm1 = torch.norm(self.w - w_dq, p=0.7)
                
                W_q, meta = Quantizer.quantize(self.w, nbits=nbits, round_zero=True, optimize=True, compute_dtype=compute_dtype, view_as_float=True)
                assert W_q.dtype == compute_dtype
                w_dq = Quantizer.dequantize(W_q, meta)
                norm2 = torch.norm(self.w - w_dq, p=0.7)
                
                self.assertTrue(torch.equal(norm1, norm2))
        
    def test_quantizer_cuda(self):
        w_cuda = self.w.cuda()
        W_q, meta = Quantizer.quantize(w_cuda, round_zero=True, optimize=True, view_as_float=False)
        w_dq = Quantizer.dequantize(W_q, meta)
        norm1 = torch.norm(w_cuda - w_dq, p=0.7)
        
        W_q, meta = Quantizer.quantize(w_cuda, round_zero=True, optimize=True, view_as_float=True)
        w_dq = Quantizer.dequantize(W_q, meta)
        norm2 = torch.norm(w_cuda - w_dq, p=0.7)
        
        self.assertTrue(torch.equal(norm1, norm2))        
        
    def test_floatview_bitpacking(self):
        shapes = [[32, 32], [256, 256], [512, 512], [1024, 1024], [2048, 2048], [4096, 4096], [8192, 8192], [8192, 4096], [8192, 128], [32, 4096]]
        for compute_dtype in [torch.float32, torch.float16, torch.bfloat16]:
            for nbits in [8, 4, 3, 2, 1]:
                for shape in shapes:
                    bit_pack      = Quantizer.bit_to_packing[nbits]
                    view_dtype    = Quantizer.unpack_view_dtype[bit_pack]
                    W             = torch.randint(0, 2**nbits, shape, device='cuda').to(view_dtype)
                    W_packed_orig = Quantizer.pack[bit_pack](W)
                    W_packed_view = W_packed_orig.clone().view(compute_dtype)
                    assert W_packed_view.dtype == compute_dtype
                    assert torch.abs(W_packed_orig - W_packed_view.view(view_dtype)).max() < 1e-5
                    
    def test_hqq_linear(self):
        for compute_dtype in [torch.float32, torch.float16, torch.bfloat16]:
            # FIXME: fails with 8-bit!
            # 8-bit norms: norm1=22.245920181274414, norm2=22.237092971801758, norm3=37.290225982666016, norm4=37.38299560546875, norm5=37.26346206665039
            for nbits in [4,3,2]:
                
                # print(f"Testing nbits={nbits} and compute_dtype={compute_dtype}")
                
                quant_configs = [
                        BaseQuantizeConfig(nbits=nbits, group_size=64, quant_zero=False, quant_scale=False, offload_meta=False, view_as_float=True),
                        BaseQuantizeConfig(nbits=nbits, group_size=64, quant_zero=True, quant_scale=False, offload_meta=False, view_as_float=True),
                        BaseQuantizeConfig(nbits=nbits, group_size=64, quant_zero=False, quant_scale=True, offload_meta=False, view_as_float=True),
                        BaseQuantizeConfig(nbits=nbits, group_size=64, quant_zero=True, quant_scale=True, offload_meta=False, view_as_float=True),
                        BaseQuantizeConfig(nbits=nbits, group_size=64, quant_zero=True, quant_scale=True, offload_meta=True, view_as_float=True),
                        BaseQuantizeConfig(nbits=nbits, group_size=64, quant_zero=False, quant_scale=False, offload_meta=True, view_as_float=True),
                ]

                w_dqs = []
                for quant_cfg in quant_configs:
                    if quant_cfg['scale_quant_params']: 
                        quant_cfg['scale_quant_params']['group_size'] = 8
                    if quant_cfg['zero_quant_params']: 
                        if quant_cfg['offload_meta']:
                            quant_cfg['zero_quant_params']['group_size'] = 8
                            quant_cfg['zero_quant_params']['channel_wise'] = True
                        else:
                            quant_cfg['zero_quant_params']['group_size'] = None
                            quant_cfg['zero_quant_params']['channel_wise'] = False
                    mq = HQQLinear(self.m, quant_cfg, compute_dtype=compute_dtype, initialize=False)
                    HQQLinear.set_backend(HQQBackend.ATEN_BACKPROP)
                    mq.initialize()
                    w_dqs.append(mq.dequantize_aten())
                
                norm1 = torch.norm(self.w.cuda() - w_dqs[0], p=0.7)
                norm2 = torch.norm(self.w.cuda() - w_dqs[1], p=0.7)
                norm3 = torch.norm(self.w.cuda() - w_dqs[2], p=0.7)
                norm4 = torch.norm(self.w.cuda() - w_dqs[3], p=0.7)
                norm5 = torch.norm(self.w.cuda() - w_dqs[4], p=0.7)
                norm6 = torch.norm(self.w.cuda() - w_dqs[5], p=0.7)
                
                # print(f"norm1={norm1}, norm2={norm2}, norm3={norm3}, norm4={norm4}, norm5={norm5}, norm6={norm6}")
                
                assert torch.isclose(norm1, norm2, rtol=1e-2)
                assert torch.isclose(norm1, norm3, rtol=1e-2)
                assert torch.isclose(norm1, norm4, rtol=1e-2)
                assert torch.isclose(norm1, norm5, rtol=1e-2)
                assert torch.isclose(norm1, norm6, rtol=1e-2)
        
    def test_hqq_linear_forward(self):
        HQQLinear.set_backend(HQQBackend.ATEN_BACKPROP)
        context_size = 4096
        batch_size   = 1
        for compute_dtype in [torch.float32, torch.float16, torch.bfloat16]:
            for nbits in [4,3,2]:
                for quant_zero in [False, True]:
                    for quant_scale in [False, True]:
                        for offload_meta in [True, False]:
                            for group_size in [8, 16, 32, 64, 128, 256]:
                                quant_config_int   = BaseQuantizeConfig(nbits=nbits, group_size=group_size, quant_zero=quant_zero, quant_scale=quant_scale, offload_meta=offload_meta, view_as_float=False)
                                quant_config_float = BaseQuantizeConfig(nbits=nbits, group_size=group_size, quant_zero=quant_zero, quant_scale=quant_scale, offload_meta=offload_meta, view_as_float=True)
                            
                                zero_scale_group_size = self.m.weight.numel() // group_size // 2 
                                if quant_config_int['scale_quant_params'] is not None: 
                                    # should be divisible by group_size
                                    quant_config_int['scale_quant_params']['group_size'] = zero_scale_group_size
                                    quant_config_float['scale_quant_params']['group_size'] = zero_scale_group_size
                                if quant_config_int['zero_quant_params'] is not None: 
                                    if quant_config_int['offload_meta']:
                                        # should be divisible by group_size                                    
                                        quant_config_int['zero_quant_params']['group_size'] = zero_scale_group_size
                                        quant_config_float['zero_quant_params']['group_size'] = zero_scale_group_size
                                        quant_config_int['zero_quant_params']['channel_wise'] = True
                                        quant_config_float['zero_quant_params']['channel_wise'] = True
                                    else:
                                        quant_config_int['zero_quant_params']['group_size'] = None
                                        quant_config_float['zero_quant_params']['group_size'] = None
                                        quant_config_int['zero_quant_params']['channel_wise'] = False
                                        quant_config_float['zero_quant_params']['channel_wise'] = False


                                hqq_linear_int     = HQQLinear(self.m, quant_config_int, compute_dtype=compute_dtype,   del_orig=False)
                                hqq_linear_float   = HQQLinear(self.m, quant_config_float, compute_dtype=compute_dtype, del_orig=False)

                                x = torch.randn([batch_size, context_size, self.m.weight.shape[1]], device='cuda').to(compute_dtype)
                                with torch.no_grad():
                                    y_int   = hqq_linear_int.forward(x)
                                    y_float = hqq_linear_float.forward(x)

                                assert torch.allclose(y_int, y_float, rtol=1e-5)
                                
    def test_save_and_load_model(self):
        compute_dtype = torch.bfloat16
        model_name = "meta-llama/Llama-2-7b-hf"
        cfg = AutoConfig.from_pretrained(model_name)
        cfg.num_hidden_layers = 2

        # load model on meta device without calling init and replace nn.Linear with Linear4bit
        model = AutoModelForCausalLM.from_config(cfg)
        
        # quantize and save
        quant_config = BaseQuantizeConfig(nbits=4, group_size=64, view_as_float=True)
        HQQModelForCausalLM.quantize_model_(model, quant_config, compute_dtype=compute_dtype)
        model.save_quantized(f"{self.tmp_dir}/models")
        
        # load model
        model_qt = HQQModelForCausalLM.from_quantized(f"{self.tmp_dir}/models")
        
        # check if the state_dicts are equal
        def assert_state_dict(v1,v2):
            if isinstance(v1, torch.Tensor):
                assert torch.isclose(v1,v2, rtol=1e-5).float().mean().item() > 0.99
            if isinstance(v1, dict):
                for _k,_v in v1.items():
                    if isinstance(_v, torch.Tensor):
                        assert torch.equal(_v, v2[_k])
                    else:
                        assert _v == v2[_k]

        for n,p in model.named_parameters():
            module_key, _, value_key = n.rpartition('.')
            d1 = model.get_submodule(module_key).state_dict()
            d2 = model_qt.get_submodule(module_key).state_dict()
            for (k1,v1),(k2,v2) in zip(d1.items(), d2.items()):
                assert k1 == k2
                assert_state_dict(v1,v2)
        
        
                        
                                    
if __name__ == '__main__':
    unittest.main()

==> hqq-0.1.8.tar.gz: hqq-0.1.8/tests/test_quantize.py <==
import unittest, tempfile
import torch
from hqq.core.quantize import Quantizer, HQQLinear, BaseQuantizeConfig, HQQBackend
from hqq.engine.hf import HQQModelForCausalLM
from transformers import AutoModelForCausalLM, AutoConfig, LlamaForCausalLM


class CustomLlamaModel(LlamaForCausalLM):
    def __init__(self, config):
        super().__init__(config)
        # replacing head with custom linear layer
        qcfg = BaseQuantizeConfig(nbits=2, group_size=64)
        linear = torch.nn.Linear(config.hidden_size, config.vocab_size, bias=False)
        self.lm_head = HQQLinear(linear, qcfg)


# python -m unittest tests.test_quantize.TestQuantizer.test_quantizer
class TestQuantizer(unittest.TestCase):

    def setUp(self) -> None:
        # set seed
        torch.manual_seed(42)
        self.m = torch.nn.Linear(16,128)
        self.w = self.m.weight.data
        # create tmp dir
        self.tmp_dir = tempfile.mkdtemp()
        return super().setUp()
    
    def tearDown(self) -> None:
        return super().tearDown()
    
    def test_quantizer(self):    
        for compute_dtype in [torch.float32, torch.float16, torch.bfloat16]:
            for nbits in [8,4,3,2,1]:
                W_q, meta = Quantizer.quantize(self.w, nbits=nbits, round_zero=True, optimize=True, view_as_float=False)
                if nbits == 3:
                    assert W_q.dtype == torch.int32
                else:
                    assert W_q.dtype == torch.uint8
                w_dq = Quantizer.dequantize(W_q, meta)
                norm1 = torch.norm(self.w - w_dq, p=0.7)
                
                W_q, meta = Quantizer.quantize(self.w, nbits=nbits, round_zero=True, optimize=True, compute_dtype=compute_dtype, view_as_float=True)
                assert W_q.dtype == compute_dtype
                w_dq = Quantizer.dequantize(W_q, meta)
                norm2 = torch.norm(self.w - w_dq, p=0.7)
                
                self.assertTrue(torch.equal(norm1, norm2))
        
    def test_quantizer_cuda(self):
        w_cuda = self.w.cuda()
        W_q, meta = Quantizer.quantize(w_cuda, round_zero=True, optimize=True, view_as_float=False)
        w_dq = Quantizer.dequantize(W_q, meta)
        norm1 = torch.norm(w_cuda - w_dq, p=0.7)
        
        W_q, meta = Quantizer.quantize(w_cuda, round_zero=True, optimize=True, view_as_float=True)
        w_dq = Quantizer.dequantize(W_q, meta)
        norm2 = torch.norm(w_cuda - w_dq, p=0.7)
        
        self.assertTrue(torch.equal(norm1, norm2))        
        
    def test_floatview_bitpacking(self):
        shapes = [[32, 32], [256, 256], [512, 512], [1024, 1024], [2048, 2048], [4096, 4096], [8192, 8192], [8192, 4096], [8192, 128], [32, 4096]]
        for compute_dtype in [torch.float32, torch.float16, torch.bfloat16]:
            for nbits in [8, 4, 3, 2, 1]:
                for shape in shapes:
                    bit_pack      = Quantizer.bit_to_packing[nbits]
                    view_dtype    = Quantizer.unpack_view_dtype[bit_pack]
                    W             = torch.randint(0, 2**nbits, shape, device='cuda').to(view_dtype)
                    W_packed_orig = Quantizer.pack[bit_pack](W)
                    W_packed_view = W_packed_orig.clone().view(compute_dtype)
                    assert W_packed_view.dtype == compute_dtype
                    assert torch.abs(W_packed_orig - W_packed_view.view(view_dtype)).max() < 1e-5
                    
    def test_hqq_linear(self):
        for compute_dtype in [torch.float32, torch.float16, torch.bfloat16]:
            # FIXME: fails with 8-bit!
            # 8-bit norms: norm1=22.245920181274414, norm2=22.237092971801758, norm3=37.290225982666016, norm4=37.38299560546875, norm5=37.26346206665039
            for nbits in [4,3,2]:
                
                # print(f"Testing nbits={nbits} and compute_dtype={compute_dtype}")
                
                quant_configs = [
                        BaseQuantizeConfig(nbits=nbits, group_size=64, quant_zero=False, quant_scale=False, offload_meta=False, view_as_float=True),
                        BaseQuantizeConfig(nbits=nbits, group_size=64, quant_zero=True, quant_scale=False, offload_meta=False, view_as_float=True),
                        BaseQuantizeConfig(nbits=nbits, group_size=64, quant_zero=False, quant_scale=True, offload_meta=False, view_as_float=True),
                        BaseQuantizeConfig(nbits=nbits, group_size=64, quant_zero=True, quant_scale=True, offload_meta=False, view_as_float=True),
                        BaseQuantizeConfig(nbits=nbits, group_size=64, quant_zero=True, quant_scale=True, offload_meta=True, view_as_float=True),
                        BaseQuantizeConfig(nbits=nbits, group_size=64, quant_zero=False, quant_scale=False, offload_meta=True, view_as_float=True),
                ]

                w_dqs = []
                for quant_cfg in quant_configs:
                    if quant_cfg['scale_quant_params']: 
                        quant_cfg['scale_quant_params']['group_size'] = 8
                    if quant_cfg['zero_quant_params']: 
                        if quant_cfg['offload_meta']:
                            quant_cfg['zero_quant_params']['group_size'] = 8
                            quant_cfg['zero_quant_params']['channel_wise'] = True
                        else:
                            quant_cfg['zero_quant_params']['group_size'] = None
                            quant_cfg['zero_quant_params']['channel_wise'] = False
                    mq = HQQLinear(self.m, quant_cfg, compute_dtype=compute_dtype, initialize=False)
                    HQQLinear.set_backend(HQQBackend.ATEN_BACKPROP)
                    mq.initialize()
                    w_dqs.append(mq.dequantize_aten())
                
                norm1 = torch.norm(self.w.cuda() - w_dqs[0], p=0.7)
                norm2 = torch.norm(self.w.cuda() - w_dqs[1], p=0.7)
                norm3 = torch.norm(self.w.cuda() - w_dqs[2], p=0.7)
                norm4 = torch.norm(self.w.cuda() - w_dqs[3], p=0.7)
                norm5 = torch.norm(self.w.cuda() - w_dqs[4], p=0.7)
                norm6 = torch.norm(self.w.cuda() - w_dqs[5], p=0.7)
                
                # print(f"norm1={norm1}, norm2={norm2}, norm3={norm3}, norm4={norm4}, norm5={norm5}, norm6={norm6}")
                
                assert torch.isclose(norm1, norm2, rtol=1e-2)
                assert torch.isclose(norm1, norm3, rtol=1e-2)
                assert torch.isclose(norm1, norm4, rtol=1e-2)
                assert torch.isclose(norm1, norm5, rtol=1e-2)
                assert torch.isclose(norm1, norm6, rtol=1e-2)
        
    def test_hqq_linear_forward(self):
        HQQLinear.set_backend(HQQBackend.ATEN_BACKPROP)
        context_size = 4096
        batch_size   = 1
        for compute_dtype in [torch.float32, torch.float16, torch.bfloat16]:
            for nbits in [4,3,2]:
                for quant_zero in [False, True]:
                    for quant_scale in [False, True]:
                        for offload_meta in [True, False]:
                            for group_size in [8, 16, 32, 64, 128, 256]:
                                quant_config_int   = BaseQuantizeConfig(nbits=nbits, group_size=group_size, quant_zero=quant_zero, quant_scale=quant_scale, offload_meta=offload_meta, view_as_float=False)
                                quant_config_float = BaseQuantizeConfig(nbits=nbits, group_size=group_size, quant_zero=quant_zero, quant_scale=quant_scale, offload_meta=offload_meta, view_as_float=True)
                            
                                zero_scale_group_size = self.m.weight.numel() // group_size // 2 
                                if quant_config_int['scale_quant_params'] is not None: 
                                    # should be divisible by group_size
                                    quant_config_int['scale_quant_params']['group_size'] = zero_scale_group_size
                                    quant_config_float['scale_quant_params']['group_size'] = zero_scale_group_size
                                if quant_config_int['zero_quant_params'] is not None: 
                                    if quant_config_int['offload_meta']:
                                        # should be divisible by group_size                                    
                                        quant_config_int['zero_quant_params']['group_size'] = zero_scale_group_size
                                        quant_config_float['zero_quant_params']['group_size'] = zero_scale_group_size
                                        quant_config_int['zero_quant_params']['channel_wise'] = True
                                        quant_config_float['zero_quant_params']['channel_wise'] = True
                                    else:
                                        quant_config_int['zero_quant_params']['group_size'] = None
                                        quant_config_float['zero_quant_params']['group_size'] = None
                                        quant_config_int['zero_quant_params']['channel_wise'] = False
                                        quant_config_float['zero_quant_params']['channel_wise'] = False


                                hqq_linear_int     = HQQLinear(self.m, quant_config_int, compute_dtype=compute_dtype,   del_orig=False)
                                hqq_linear_float   = HQQLinear(self.m, quant_config_float, compute_dtype=compute_dtype, del_orig=False)

                                x = torch.randn([batch_size, context_size, self.m.weight.shape[1]], device='cuda').to(compute_dtype)
                                with torch.no_grad():
                                    y_int   = hqq_linear_int.forward(x)
                                    y_float = hqq_linear_float.forward(x)

                                assert torch.allclose(y_int, y_float, rtol=1e-5)

    @staticmethod
    def assert_equal_models(model, model_qt):
        def assert_state_dict(v1, v2):
            if isinstance(v1, torch.Tensor):
                assert torch.isclose(v1, v2, rtol=1e-5).float().mean().item() > 0.99
            if isinstance(v1, dict):
                for _k, _v in v1.items():
                    if isinstance(_v, torch.Tensor):
                        assert torch.equal(_v, v2[_k])
                    else:
                        assert _v == v2[_k]

        for n, p in model.named_parameters():
            module_key, _, value_key = n.rpartition('.')
            d1 = model.get_submodule(module_key).state_dict()
            d2 = model_qt.get_submodule(module_key).state_dict()
            for (k1, v1), (k2, v2) in zip(d1.items(), d2.items()):
                assert k1 == k2
                assert_state_dict(v1, v2)

    def test_save_and_load_model(self):
        compute_dtype = torch.bfloat16
        model_name = "meta-llama/Llama-2-7b-hf"
        cfg = AutoConfig.from_pretrained(model_name)
        cfg.num_hidden_layers = 2

        # load model on meta device without calling init and replace nn.Linear with Linear4bit
        model = AutoModelForCausalLM.from_config(cfg)

        # quantize and save
        quant_config = BaseQuantizeConfig(nbits=4, group_size=64, view_as_float=True)
        HQQModelForCausalLM.quantize_model_(model, quant_config, compute_dtype=compute_dtype)
        model.save_quantized(f"{self.tmp_dir}/models")

        # load model
        model_qt = HQQModelForCausalLM.from_quantized(f"{self.tmp_dir}/models", compute_dtype=compute_dtype)

        # check if the state_dicts are equal
        self.assert_equal_models(model, model_qt)

    def test_create_and_load_custom_model_with_hqqlinear_from_state_dict(self):
        model_name = "meta-llama/Llama-2-7b-hf"
        cfg = AutoConfig.from_pretrained(model_name)
        cfg.num_hidden_layers = 2

        model = CustomLlamaModel(cfg)
        state_dict = model.state_dict()

        loaded_model = CustomLlamaModel(cfg)
        loaded_model.load_state_dict(state_dict)

        # check if the state_dicts are equal
        self.assert_equal_models(model, loaded_model)


if __name__ == '__main__':
    unittest.main()
